from __future__ import annotations

import os
from ctypes import *
from datetime import datetime
from os.path import join as jpath
from typing import Sequence, Literal, Annotated

import numpy as np
from numpy.ctypeslib import as_ctypes

from .config import IRI_VERSIONS, DEFAULT_IRI_VERSION
from .data_update import update
from .iri_flags import get_jf
from .irioutput import IRIOutput
from .read_iri_data import read_apf107

_iri_cfd = os.path.dirname(os.path.abspath(__file__))

_APF107_DATA, _LAST_DATE = read_apf107()


try:
    iri2016 = np.ctypeslib.load_library("libiri2016", _iri_cfd)
    iri2020 = np.ctypeslib.load_library("libiri2020", _iri_cfd)
except OSError:
    raise ImportError(
        """
        Could not import IRI libraries, most probably they were not compiled during installation. You can try 
        compiling them manually:
        1) Locate the package installation directory and open it in terminal
        2) Run: 
            >>> cmake .
            >>> make
        """
    )


def indices_uptodate(dt: datetime):
    global _APF107_DATA, _LAST_DATE
    if dt > _LAST_DATE:
        print("Requested date is not covered in database. Updating indices.")
        if update():
            _APF107_DATA, _LAST_DATE = read_apf107()
            if dt > _LAST_DATE:
                raise RuntimeError("Requested indices are not available yet. The update latency is usually 2-3 days.")
        else:
            raise RuntimeError("Cannot update indices. Check internet connection.")


def iri(dt: datetime, altrange: Annotated[Sequence[float], 3], lat: float | Sequence[float],
        lon: float | Sequence[float], version: Literal[16, 20] = DEFAULT_IRI_VERSION,
        jf: np.ndarray | str = None, **kwargs) -> IRIOutput:
    """
    The main function of the package representing the IRI_SUB routine from the IRI source code. Provides access
    to all IRI calculations through the wide range of parameters.

    :param dt: Date and time - must be a single value.
    :param altrange: Range of altitudes in the form (alt_start, alt_stop, alt_step), all in [km].
    :param lat: Geographical latitude.
    :param lon: Geographical longitude.
    :param version: Version of the IRI to use:

                    * ``version=16`` - IRI-2016
                    * ``version=20`` - IRI-2020

    :param jf: Array of JF parameters or string for predefined JF arrays to be used in the IRI_SUB function. See
               :func:`iricore.get_jf` for details. If not specified otherwise, the default IRI JF array will be used.
    :param kwargs: Use it to enter user input values for the IRI calculation. Names of parameters are
                   'oarr'+'[oarr index you want to modify]'. Available parameters are:

                   * **oarr0** - float - user input for foF2/MHz or NmF2/m-3
                   * **oarr1** - float - user input for hmF2/km or M(3000)F2
                   * **oarr2** - float - user input for foF1/MHz or NmF1/m-3
                   * **oarr3** - float - user input for hmF1/km
                   * **oarr4** - float - user input for foE/MHz or NmE/m-3
                   * **oarr5** - float - user input for hmE/km
                   * **oarr9** - float - user input for B0
                   * **oarr14** - (float, float) - user input for Ne(300km), Ne(400km)/m-3. Use oarr14[...]=-1 if one of
                     these values is not available. If jf(22)==False then Ne(300km), Ne(550km)/m-3.
                   * **oarr32** - float - user input for Rz12
                   * **oarr34** - float - user input for B1
                   * **oarr38** - float - user input for IG12
                   * **oarr40** - float - user input for daily F10.7 index (make sure to also specify oarr45, otherwise it
                     will be copied from oarr40)
                   * **oarr45** - float - user input for 81-day avg F10.7

    :return: :class:`iricore.IRIOutput` dataclass.
    """
    indices_uptodate(dt)
    lat = np.atleast_1d(np.asarray(lat))
    lon = np.atleast_1d(np.asarray(lon))
    if not len(lat) == len(lon):
        raise ValueError("Lengths of latitude and longitude arrays must be equal.")

    if not isinstance(jf, np.ndarray):
        jf = jf or "default"
        if isinstance(jf, str):
            jf = get_jf(jf)
    else:
        if isinstance(jf, np.ndarray) and jf.size != 50:
            raise ValueError("Length of jf array must be 50")

    if version == 16:
        iricore = iri2016
    elif version == 20:
        iricore = iri2020
    else:
        raise ValueError(f"Available IRI versions: " + ", ".join(IRI_VERSIONS))

    if (altrange[1] - altrange[0]) / altrange[2] + 1 > 1000:
        raise ValueError("The specified altitude range and step require more that 1000 points for calculation, "
                         "which exceeds IRI limitations. Please consider breaking calculation into chunks.")

    # Converting parameters to Fortran types
    jmag = False
    iyyyy = c_int(dt.year)
    mmdd = c_int(100 * dt.month + dt.day)
    dhour = c_float(dt.hour + dt.minute / 60 + dt.second / 3600 + 25.)
    glat = as_ctypes(np.array(lat, dtype=np.float32, order="F"))
    glon = as_ctypes(np.array(lon, dtype=np.float32, order="F"))
    gsize = c_int(len(lat))
    heibeg = c_float(altrange[0])
    heiend = c_float(altrange[1])
    heistp = c_float(altrange[2])
    oarr_out = np.zeros((100, len(lat)), dtype=np.float32, order="F")
    outf_out = np.zeros((20, 1000, len(lat)), dtype=np.float32, order='F')

    # Handling manual user input
    if len(kwargs) > 0:
        if "oarr0" in kwargs.keys():
            jf[7] = 0
            oarr_out[0, :] = kwargs["oarr0"]
        if "oarr1" in kwargs.keys():
            jf[8] = 0
            oarr_out[1, :] = kwargs["oarr1"]
        if "oarr14" in kwargs.keys():
            jf[9] = 0
            oarr_out[14, :] = kwargs["oarr14"][0]
            oarr_out[15, :] = kwargs["oarr14"][1]
        if "oarr2" in kwargs.keys():
            jf[12] = 0
            oarr_out[2, :] = kwargs["oarr2"]
        if "oarr3" in kwargs.keys():
            jf[13] = 0
            oarr_out[3, :] = kwargs["oarr3"]
        if "oarr4" in kwargs.keys():
            jf[14] = 0
            oarr_out[4, :] = kwargs["oarr4"]
        if "oarr5" in kwargs.keys():
            jf[15] = 0
            oarr_out[5, :] = kwargs["oarr5"]
        if "oarr32" in kwargs.keys():
            jf[16] = 0
            oarr_out[32, :] = kwargs["oarr32"]
        if "oarr40" in kwargs.keys():
            jf[24] = 0
            oarr_out[40, :] = kwargs["oarr40"]
        if "oarr38" in kwargs.keys():
            jf[26] = 0
            oarr_out[38, :] = kwargs["oarr38"]
        if "oarr45" in kwargs.keys():
            jf[31] = 0
            oarr_out[45, :] = kwargs["oarr45"]
        if "oarr9" in kwargs.keys():
            jf[42] = 0
            oarr_out[9, :] = kwargs["oarr9"]
        if "oarr34" in kwargs.keys():
            jf[43] = 0
            oarr_out[34, :] = kwargs["oarr34"]

    datadir = jpath(_iri_cfd, 'data')
    datadir_bytes = bytes(datadir, 'utf-8')
    aap, af107, nlines = _APF107_DATA

    # Calling IRI_SUB
    iricore.iricore_(as_ctypes(jf), byref(c_bool(jmag)), glat, glon, byref(gsize),
                     byref(iyyyy), byref(mmdd), byref(dhour),
                     byref(heibeg), byref(heiend), byref(heistp),
                     # oarr_out.ctypes.data_as(POINTER(c_float)),
                     oarr_out.ctypes.data_as(POINTER(c_float)),
                     outf_out.ctypes.data_as(POINTER(c_float)),
                     datadir_bytes, byref(c_int(len(datadir))),
                     aap.ctypes.data_as(POINTER(c_float)), af107.ctypes.data_as(POINTER(c_float)), byref(c_int(nlines))
                     )

    outf_out = np.ascontiguousarray(outf_out)
    oarr_out = np.ascontiguousarray(oarr_out)
    return IRIOutput.from_raw(outf_out, oarr_out, lat, lon, altrange)



from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class IRIOutput:
    """
    Dataclass containing :func:`iricore.iri` output. If lat/lon are 1D arrays, then the
    IRI returned values will a 2D array with shape (lat.size, alt.size). If lat/lon
    are a single number, the returned values will be an 1D array with shape (alt.size).

    :param lat: Array of latitudes corresponding to the IRI output (0 axis).
    :param lon: Array of longitudes corresponding to the IRI output (0 axis).
    :param height: Array of heights corresponding to the IRI output (1 axis).
    :param edens: Electron density in [m-3].
    :param ntemp: Neutral temperature in [K].
    :param itemp: Ion temperature in [K].
    :param etemp: Electron temperature in [K].
    :param o: O+ ion density in [%](default) or [m-3].
    :param h: H+ ion density in [%](default) or [m-3].
    :param he: He+ ion density in [%](default) or [m-3].
    :param o2: O2+ ion density in [%](default) or [m-3].
    :param no: NO+ ion density in [%](default) or [m-3].
    :param cluster: Cluster ion density in [%](default) or [m-3].
    :param n: N+ ion density in [%](default) or [m-3].
    :param firi_output: If FIRI was manually selected for the D-region, the ``firi_output`` will be an 1D array
                        containing:

                        * [0:10] standard IRI-Ne for 60,65,..,110km;
                        * [11:21] Friedrich (FIRI) model at these heights;
                        * [22:32] standard Danilov (SW=0, WA=0);
                        * [33:43] for minor Stratospheric Warming (SW=0.5);
                        * [55:65] weak Winter Anomaly (WA=0.5) conditions;
                        * [44:54] for major Stratospheric Warming (SW=1);
                        * [66:76] strong Winter Anomaly (WA=1) conditions.
    :param oarr: Additional output data. First index axis corresponds to different coordinates, for example,
                 oarr[0] will correspond to lat[0] and lon[0] and so on. The second index axis has size 100 and maps
                 the extra output information. If only one coordinate was calculated, the first axis is discarded.

                        * oarr[0] = NMF2/M-3
                        * oarr[1] = HMF2/KM
                        * oarr[2] = NMF1/M-3
                        * oarr[3] = HMF1/KM
                        * oarr[4] = NME/M-3
                        * oarr[5] = HME/KM
                        * oarr[6] = NMD/M-3
                        * oarr[7] = HMD/KM
                        * oarr[8] = HHALF/KM
                        * oarr[9] = B0/KM
                        * oarr[10] = VALLEY-BASE/M-3
                        * oarr[11] = VALLEY-TOP/KM
                        * oarr[12] = TE-PEAK/K
                        * oarr[13] = TE-PEAK HEIGHT/KM
                        * oarr[14] = TE-MOD(300KM)
                        * oarr[15] = TE-MOD(400KM)/K
                        * oarr[16] = TE-MOD(600KM)
                        * oarr[17] = TE-MOD(1400KM)/K
                        * oarr[18] = TE-MOD(3000KM)
                        * oarr[19] = TE(120KM)=TN=TI/K
                        * oarr[20] = TI-MOD(430KM)
                        * oarr[21] = X/KM, WHERE TE=TI
                        * oarr[22] = SOL ZENITH ANG/DEG
                        * oarr[23] = SUN DECLINATION/DEG
                        * oarr[24] = DIP/deg
                        * oarr[25] = DIP LATITUDE/deg
                        * oarr[26] = MODIFIED DIP LAT.
                        * oarr[27] = Geographic latitude
                        * oarr[28] = sunrise/dec. hours
                        * oarr[29] = sunset/dec. hours
                        * oarr[30] = ISEASON (1=spring)
                        * oarr[31] = Geographic longitude
                        * oarr[32] = Rz12
                        * oarr[33] = Covington Index
                        * oarr[34] = B1
                        * oarr[35] = M(3000)F2
                        * oarr[36] = TEC/m-2
                        * oarr[37] = TEC_top/TEC*100.
                        * oarr[38] = gind (IG12)
                        * oarr[39] = F1 probability
                        * oarr[40] = F10.7 daily
                        * oarr[41] = c1 (F1 shape)
                        * oarr[42] = daynr
                        * oarr[43] = equatorial vertical ion drift in m/s
                        * oarr[44] = foF2_storm/foF2_quiet
                        * oarr[45] = F10.7_81
                        * oarr[46] = foE_storm/foE_quiet
                        * oarr[47] = spread-F probability
                        * oarr[48] = Geomag. latitude
                        * oarr[49] = Geomag. longitude
                        * oarr[50] = ap at current time
                        * oarr[51] = daily ap
                        * oarr[52] = invdip/degree
                        * oarr[53] = MLT-Te
                        * oarr[54] = CGM-latitude
                        * oarr[55] = CGM-longitude
                        * oarr[56] = CGM-MLT
                        * oarr[57] = CGM lat eq. aurl bodry
                        * oarr[58] = CGM-lati(MLT=0)
                        * oarr[59] = CGM-lati for MLT=1
                        * oarr[60] = CGM-lati(MLT=2)
                        * oarr[61] = CGM-lati for MLT=3
                        * oarr[62] = CGM-lati(MLT=4)
                        * oarr[63] = CGM-lati for MLT=5
                        * oarr[64] = CGM-lati(MLT=6)
                        * oarr[65] = CGM-lati for MLT=7
                        * oarr[66] = CGM-lati(MLT=8)
                        * oarr[67] = CGM-lati for MLT=9
                        * oarr[68] = CGM-lati(MLT=10)
                        * oarr[69] = CGM-lati for MLT=11
                        * oarr[70] = CGM-lati(MLT=12)
                        * oarr[71] = CGM-lati for MLT=13
                        * oarr[72] = CGM-lati(MLT=14)
                        * oarr[73] = CGM-lati for MLT=15
                        * oarr[74] = CGM-lati(MLT=16)
                        * oarr[75] = CGM-lati for MLT=17
                        * oarr[76] = CGM-lati(MLT=18)
                        * oarr[77] = CGM-lati for MLT=19
                        * oarr[78] = CGM-lati(MLT=20)
                        * oarr[79] = CGM-lati for MLT=21
                        * oarr[80] = CGM-lati(MLT=22)
                        * oarr[81] = CGM-lati for MLT=23
                        * oarr[82] = Kp at current time
                        * oarr[83] = magnetic declination
                        * oarr[84] = L-value
                        * oarr[85] = dipole moment
                        * oarr[86] = SAX300
                        * oarr[87] = SUX300
                        * oarr[88] = HNEA
                        * oarr[89] = HNEE
    """
    lat: np.ndarray
    lon: np.ndarray
    height: np.ndarray
    edens: np.ndarray
    ntemp: np.ndarray
    itemp: np.ndarray
    etemp: np.ndarray
    o: np.ndarray
    h: np.ndarray
    he: np.ndarray
    o2: np.ndarray
    no: np.ndarray
    cluster: np.ndarray
    n: np.ndarray
    firi_output: np.ndarray
    oarr: np.ndarray

    @staticmethod
    def _extract_iri_row(row, arr_shape):
        ncoord, nalt = arr_shape
        row = row.transpose().reshape((ncoord, -1))[:, :nalt]
        return np.squeeze(np.where(row < 0, np.nan, row))

    @classmethod
    def from_raw(cls, outf_out: np.ndarray, oarr_out: np.ndarray, lat: np.ndarray, lon: np.ndarray,
                 height_range: [float, float, float]):
        height = np.arange(height_range[0], height_range[1] + height_range[2] / 2, height_range[2])
        arr_shape = (lat.size, height.size)
        return cls(
            lat=lat,
            lon=lon,
            height=height,
            edens=cls._extract_iri_row(outf_out[0], arr_shape),
            ntemp=cls._extract_iri_row(outf_out[1], arr_shape),
            itemp=cls._extract_iri_row(outf_out[2], arr_shape),
            etemp=cls._extract_iri_row(outf_out[3], arr_shape),
            o=cls._extract_iri_row(outf_out[4], arr_shape),
            h=cls._extract_iri_row(outf_out[5], arr_shape),
            he=cls._extract_iri_row(outf_out[6], arr_shape),
            o2=cls._extract_iri_row(outf_out[7], arr_shape),
            no=cls._extract_iri_row(outf_out[8], arr_shape),
            cluster=cls._extract_iri_row(outf_out[9], arr_shape),
            n=cls._extract_iri_row(outf_out[10], arr_shape),
            firi_output=cls._extract_iri_row(outf_out[13], arr_shape),
            oarr=np.squeeze(oarr_out.swapaxes(0, 1))
        )



import os
from datetime import datetime

import fortranformat as ff
import numpy as np

_iri_cfd = os.path.dirname(os.path.abspath(__file__))
_datadir = os.path.join(_iri_cfd, f"data/")


def read_apf107():
    with open(os.path.join(_datadir, "index/apf107.dat"), "r") as file:
        lines = file.readlines()
    nlines = len(lines)
    linereader = ff.FortranRecordReader("(13I3, 3F5.1)")
    data = np.array([linereader.read(line_) for line_ in lines], order='F', dtype=np.float32)

    f107d = data[:, 13]
    f107_81 = data[:, 14]
    f107_365 = data[:, 15]
    af107 = np.zeros((27000, 3), order='F', dtype=np.float32)
    af107[:nlines, 0] = f107d
    af107[:nlines, 1] = np.where(f107_81 < -4, f107d, f107_81)
    af107[:nlines, 2] = np.where(f107_365 < -4, f107d, f107_365)
    aap = np.zeros((27000, 9), order='F', dtype=np.int32)
    aap[:nlines, :] = data[:, 3:12].astype(np.int32)
    ly, lm, ld = data[nlines - 1, :3].astype(np.int8)
    last_date = datetime(ly + 2000, lm, ld)
    return (aap, af107, nlines), last_date



# A simple 2D raytracing. Code copied from dionpy

from __future__ import annotations

import datetime
from typing import Tuple

import numpy as np
import pymap3d as pm
from pymap3d import Ellipsoid

from .iri import iri
from .modules.ion_tools import srange, refr_index, refr_angle, plasfreq, R_EARTH

_ROUND_ELL = Ellipsoid(R_EARTH, R_EARTH)
_LIGHT_SPEED = 2.99792458e8  # in [m/s]


def _raytrace_sublayer(lat_ray, lon_ray, h_ray, h_next, alt_cur, az, freq, d_theta, ref_ind, n_sublayer, nlayers, dt,
                       theta_ref=None):
    # Distance from current position to next layer
    r_slant = srange(np.deg2rad(90 - alt_cur), h_next - h_ray, re=R_EARTH + h_ray)
    lat_next, lon_next, _ = pm.aer2geodetic(az, alt_cur, r_slant, lat_ray, lon_ray, h_ray, ell=_ROUND_ELL)
    # The sides of the 1st triangle
    d_cur = R_EARTH + h_ray  # Distance from Earth center to current point
    d_next = R_EARTH + h_next  # Distance from Earth center to layer

    if theta_ref is None:
        # The inclination angle at the interface using law of cosines [rad]
        costheta_inc = (r_slant ** 2 + d_next ** 2 - d_cur ** 2) / (2 * r_slant * d_next)
        assert (costheta_inc <= 1).all(), (f"Cosine of inclination angle cannot be >= 1. Something is wrong with "
                                           f"coordinates at heights {h_ray * 1e-3:.1f}-{h_next * 1e-3:.1f} [km].")
        theta_inc = np.arccos(costheta_inc)
    else:
        # Angle between d_cur and r_slant
        int_angle_rad = np.pi - theta_ref
        # The inclination angle at the i-th interface using law of sines [rad]
        theta_inc = np.arcsin(np.sin(int_angle_rad) * d_cur / d_next)

    # Get IRI info of point
    ed = iri(dt, [h_next / 1e3, h_next / 1e3, 1], lat_next, lon_next).edens
    ed = np.where(ed < 0, 0, ed)

    # Refraction index of the surface
    if n_sublayer == nlayers - 1:
        ref_ind_next = np.ones(alt_cur.shape)
        nan_theta_mask = np.zeros(alt_cur.shape)
    else:
        ref_ind_next = refr_index(ed, freq)
        nan_theta_mask = plasfreq(ed, angular=False) > freq

    # The outgoing angle at the 1st interface using Snell's law
    theta_ref = refr_angle(ref_ind, ref_ind_next, theta_inc)
    inf_theta_mask = np.abs((ref_ind / ref_ind_next * np.sin(theta_inc))) > 1
    d_theta += theta_ref - theta_inc
    alt_next = np.rad2deg(np.pi / 2 - theta_ref)
    return lat_next, lon_next, h_next, d_theta, alt_next, ref_ind_next, theta_ref, ed, nan_theta_mask, inf_theta_mask, r_slant


def raytrace(
        alt: float | np.ndarray,
        az: float | np.ndarray,
        freq: float | np.ndarray,
        pos: Tuple[float, float, float],
        heights: np.ndarray,
        dt: datetime.datetime
) -> dict:
    # Initialization of variables
    freq *= 1e6
    alt_cur = np.array(alt)
    az = np.array(az)
    heights = heights * 1e3  # in [m]
    nlayers = heights.size

    delta_theta = 0 * alt_cur
    delta_theta_hist = np.empty((*alt_cur.shape, nlayers))
    inf_theta_mask = 0 * alt_cur
    nan_theta_mask = 0 * alt_cur

    history = {
        'edens': np.zeros((*alt_cur.shape, nlayers)),
        'lat': np.zeros((*alt_cur.shape, nlayers)),
        'lon': np.zeros((*alt_cur.shape, nlayers)),
        'h': np.zeros((*alt_cur.shape, nlayers)),
        'ds': np.zeros((*alt_cur.shape, nlayers))
    }

    # Init values for the first sub-layer
    ref_ind_cur = np.ones(alt_cur.shape)
    lat_ray, lon_ray, h_ray = pos
    theta_ref = None
    for i in range(nlayers):
        # Tracing change in position due to refraction
        (lat_ray, lon_ray, h_ray, delta_theta, alt_cur, ref_ind_cur, theta_ref, ed,
         nt_mask, it_mask, ds) = _raytrace_sublayer(
            lat_ray, lon_ray, h_ray, heights[i], alt_cur, az, freq, delta_theta, ref_ind_cur, i, nlayers, dt, theta_ref)
        delta_theta_hist[..., i] = delta_theta
        nan_theta_mask += nt_mask
        inf_theta_mask += it_mask
        history['edens'][..., i] = ed
        history['lat'][..., i] = lat_ray
        history['lon'][..., i] = lon_ray
        history['h'][..., i] = h_ray / 1e3
        history['ds'][..., i] = ds / 1e3
        history['lat'][..., i] = np.where(inf_theta_mask == 0, history['lat'][..., i], np.inf)
        history['lon'][..., i] = np.where(inf_theta_mask == 0, history['lon'][..., i], np.inf)
        history['lat'][..., i] = np.where(nan_theta_mask == 0, history['lat'][..., i], np.nan)
        history['lon'][..., i] = np.where(nan_theta_mask == 0, history['lon'][..., i], np.nan)

        if inf_theta_mask:
            history['lat'][..., i:] = np.inf
            history['lon'][..., i:] = np.inf
            history['edens'][..., i] = np.nan
            history['h'][..., i] = np.nan
            history['ds'][..., i] = np.nan
            return history

    return history


def raytrace_star(args):
    """
    For parallel calculations
    """
    return raytrace(*args)
